"use strict";
import MySQL from "./MySQL.js";
import PgSQL from "./PgSQL.js";
import MongoDB from "./MongoDB.js";

class AppDatasource {
    constructor(dbConfig = null, mode = "production") {
        if (AppDatasource.instance) {
            // Returns the appDatasource instance if it's already been created before.
            return AppDatasource.instance;
        }

        if (!dbConfig) {
            throw new Error("dbConfig is required to initialize AppDatasource.");
        }

        this.dbConfig = mode === "test" ? dbConfig.test : dbConfig.default;
        this.dataSource = this.dbConfig.dataSource;
        this.initDriver();
        AppDatasource.instance = this;
        return this;
    }    

    initDriver() {
        if (this.dataSource === "MySQL") {
            this.mysql = new MySQL(this.dbConfig);
            this.mysql.createPool();
        } else if (this.dataSource === "PgSQL") {
            this.pgsql = new PgSQL(this.dbConfig);
            this.pgsql.createPool();
        } else if (this.dataSource === "MongoDB") {
            this.mongoDB = new MongoDB(this.dbConfig);
            this.mongoDB.createPool();
        } else if (dataSource) {
            throw new Error(`Datasource ${this.dataSource} not found.`);
        }
    }

    delete(table, conditions, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.delete(table, conditions, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.delete(table, conditions, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.delete(table, conditions);
        }

        return null;
    }

    deleteMany(table, conditions) {
        if (this.dataSource === "MongoDB") {
            return this.mongoDB.deleteMany(table, conditions);
        }

        return null;
    }

    save(table, data, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.save(table, data, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.save(table, data, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.save(table, data);
        }

        return null;
    }

    saveMany(table, data, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.saveMany(table, data, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.saveMany(table, data, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.saveMany(table, data);
        }

        return null;
    }

    update(table, data, conditions, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.update(table, data, conditions, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.update(table, data, conditions, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.update(table, data, conditions);
        }

        return null;
    }

    updateMany(table, data, conditions, connection = null) {
        if (this.dataSource === "MySQL") {
            // TODO: Implementar método updateMany para MySQL.
            // return this.mysql.updateMany(table, data, connection);
            throw new Error("Update many was not implemented for MySQL.");
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.updateMany(table, data, conditions, connection);
        } else if (this.dataSource === "MongoDB") {
            // TODO: Implementar método updateMany para MongoDB.
            // return this.mongoDB.updateMany(table, data);
            throw new Error("Update many was not implemented for mongodb.");
        }

        return null;
    }

    findBy(table, params, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.findBy(table, params, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.findBy(table, params, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.findBy(table, params);
        }

        return null;
    }

    findAll(table, params, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.findAll(table, params, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.findAll(table, params, connection);
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB.findAll(table, params);
        }

        return null;
    }

    find(table, type, params, connection = null) {
        if (this.dataSource === "MySQL") {
            return this.mysql.find(table, type, params, connection);
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql.find(table, type, params, connection);
        }

        return null;
    }

    showSQL() {
        if (this.dataSource === "MySQL") {
            this.mysql.showSQL();
        } else if (this.dataSource === "PgSQL") {
            this.pgsql.showSQL();
        }
    }

    getDatasource() {
        if (this.dataSource === "MySQL") {
            return this.mysql;
        } else if (this.dataSource === "PgSQL") {
            return this.pgsql;
        } else if (this.dataSource === "MongoDB") {
            return this.mongoDB;
        }

        return null;
    }

}

export default AppDatasource;
