"use strict";

import Logger from "../Component/Logger.js";
import pkg from "formidable";
const { IncomingForm } = pkg;

let logger = new Logger();

class FactoryController {
    constructor(application, mode) {
        this.application = application;
        this.core = {};
        this.mode = mode;
    }

    async init() {
        await this.loadCoreConfig();
    }

    async loadCoreConfig() {
        const coreConfigPath = `${this.application.configPath}/core.json`;
        const coreConfigModule = await import(coreConfigPath, { assert: { type: 'json' } });
        this.core = coreConfigModule.default; // O JSON será importado como um objeto
    }    

    /**
     * It's returns a config file.
     * @param {string} name Configuration's name.
     * @param {Object} app Application.
     * @returns {Object} Object configuration.
     */
    async config(name, app) {
        const importConfig = await import(`${app.configPath}/${name}`, { assert: { type: "json" } });        

        return importConfig.default;
    }

    /**
     * It's returns a new controller instance.
     * @param {string} name Controller's name.
     * @returns {Object} Object Controller.
     */
    async controller(name) {
        const that = this;
        let Controller = null;
        try {
            Controller = await import(`${this.application.controllersPath}/${name}Controller.js`);
        } catch (err) {
            if (err.code === "MODULE_NOT_FOUND") {
                throw new Error("RESOURCE_NOT_FOUND");
            }

            throw err;
        }
        
        const instance = new Controller.default();

        instance.config = function (configName) {
            return that.config(configName, that.application);
        };

        instance.model = async function (modelName) {
            if (Array.isArray(modelName)) {
                for (let index = 0; index < modelName.length; index += 1) {
                    const strModelName = modelName[index];
                    instance[strModelName] = await that.model(strModelName);
                }
            } else {
                instance[modelName] = await that.model(modelName);
            }
        };

        instance.setHeader = function (header) {
            instance.responseHeader = header;
        };

        instance.getFormData = function (request, path = null) {
            if (!request) {
                throw new Error("Param request is required.");
            }

            return new Promise((resolve, reject) => {
                const form = new IncomingForm();
                form.keepExtensions = true;
                form.multiples = true;
                form.maxFileSize = Reflect.has(this.core, "maxFileSize") ? this.core.maxFileSize : 200 * 1024 * 1024;

                if (path) {
                    form.uploadDir = path;
                } else if (Reflect.has(this.core, "uploadDir")) {
                    form.uploadDir = this.core.uploadDir;
                } else {
                    reject(new Error("Please, sets the uploadDir in core.json"));

                    return;
                }

                form.parse(request, (err, fields, files) => {
                    if (err) {
                        logger.error(`[Magic] ${JSON.stringify(err, null, 4)}`);
                        request.connection.destroy();
                        reject();

                        return;
                    }

                    const payload = {
                        data: fields,
                        files: files,
                    };
                    logger.info(`[Magic] FormData ${JSON.stringify(payload, null, 4)}`);
                    resolve(payload);
                });
            });
        };

        return instance;
    }

    /**
     * It's returns a new model instance.
     * @param {string} name Model's name.
     * @returns {Object} Object model.
     */
    async model(name) {
        const that = this;
        const Model = await import(`${this.application.modelPath}/${name}.js`);        
        const instance = new Model.default();

        instance.config = function (configName) {
            return that.config(configName, that.application);
        };

        if (typeof instance.init !== "undefined") {
            instance.init();
        }

        return instance;
    }
}

export default FactoryController;
