/*eslint max-depth: ["error", 10] max-statements: ["error", 30]*/
"use strict";

import AppDatasource from "../Component/DataSource/AppDatasource.js";

class AppModel {
    constructor() {
        this.appDatasource = new AppDatasource();
    }

    /**
     * Rules to be validated before invoke save, saveMany or update method.
     * @param {Array} arrayRules - List with rules.
     * @returns {Void} nothing.
     */
    validate(arrayRules) {
        this._arrayRules = arrayRules;
    }

    join(data) {
        this._join = data;
    }

    /**
     * This method is used to check the validation rules. It's function is invoked ever before save, saveMany or update method.
     * @param {Object} data Object to save.
     * @returns {Void} .
     */
    _checkRules(data) {
        /**
         * Start the validations
         * @param {string|number} value - Value to test
         * @param {Object} objRule - Rule that will be used to validate the value parameter.
         * @returns {Void} nothing
         */
        const process = (value, objRule) => {
            if (objRule.rule === "notEmpty" && (typeof value === "undefined" || value === "")) {
                throw new Error(objRule.message);
            } else if (objRule.rule === "maxLength" && value.length > objRule.value) {
                throw new Error(objRule.message);
            } else if (objRule.rule === "minLength" && value.length < objRule.value) {
                throw new Error(objRule.message);
            } else if (objRule.rule === "minValue" && value < objRule.value) {
                throw new Error(objRule.message);
            }
        };

        /**
         * Its checks which property are necessary to each rule.
         * @param {Object} objRule - The developer configuration rules.
         * @returns {Void} nothing
         */
        const validateRule = objRule => {
            //Sample object: { field1: { rule: 'notEmpty', message: 'Required field field1.' } }
            if (objRule.rule === "maxLength" || objRule.rule === "minLength" || objRule.rule === "minValue") {
                if (!Reflect.has(objRule, "value")) {
                    throw new Error(`The ${objRule.rule} rule requires the 'value' property.`);
                } else if (!Reflect.has(objRule, "message")) {
                    throw new Error(`The ${objRule.rule} rule requires the 'message' property.`);
                }
            } else if (!Reflect.has(objRule, "message")) {
                throw new Error(`The ${objRule.rule} rule requires the 'message' property.`);
            }
        };

        if (typeof this._arrayRules !== "undefined" && this._arrayRules.length > 0) {
            for (const key in this._arrayRules) {
                if (Reflect.has(this._arrayRules, key)) {
                    const objRule = this._arrayRules[key];
                    const fieldRule = Reflect.ownKeys(objRule)[0];

                    if (Array.isArray(data)) {
                        for (let index = 0; index < data.length; index += 1) {
                            if (!Reflect.has(data[index], fieldRule)) {
                                throw new Error(`Field ${fieldRule} not found.`);
                            }
                        }
                    } else if (!Reflect.has(data, fieldRule)) {
                        throw new Error(`Field ${fieldRule} not found.`);
                    }

                    if (Array.isArray(objRule[fieldRule])) {
                        for (let index = 0; index < objRule[fieldRule].length; index += 1) {
                            validateRule(objRule[fieldRule][index]);
                            process(data[fieldRule], objRule[fieldRule][index]);
                        }
                    } else {
                        validateRule(objRule[fieldRule]);
                        //Check all rules.
                        if (Array.isArray(data)) {
                            for (let index = 0; index < data.length; index += 1) {
                                process(data[index][fieldRule], objRule[fieldRule]);
                            }
                        } else {
                            process(data[fieldRule], objRule[fieldRule]);
                        }
                    }
                }
            }
        }

        return Promise.resolve();
    }

    /**
     * Use this to delete a record from database.
     * @param {Object} conditions - Object conditions.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of delete method.
     */
    delete(conditions, connection = null) {
        return this.appDatasource.delete(this.useTable, conditions, connection);
    }

    /**
     * Use this to delete many records from database.
     * @param {Object} conditions - Object conditions.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of delete method.
     */
    deleteMany(conditions, connection = null) {
        return this.appDatasource.deleteMany(this.useTable, conditions, connection);
    }

    /**
     * Use this to insert a record in database.
     * @param {Object} data - Object data to be inserted.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of save method.
     */
    save(data, connection = null) {
        return this._checkRules(data)
            .then(() => {
                if (typeof this.beforeSave === "function") {
                    return this.beforeSave(data);
                }

                return Promise.resolve(data);
            })
            .then(values => this.appDatasource.save(this.useTable, values, connection));
    }

    /**
     * Use this to insert many records in database.
     * @param {Array} data - List of data to be inserted.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of saveMany method.
     */
    saveMany(data, connection = null) {
        return this._checkRules(data).then(() => this.appDatasource.saveMany(this.useTable, data, connection));
    }

    /**
     * Use this to update a record in database.
     * @param {Object} data - Object data to be updated.
     * @param {Object} conditions - Object conditions.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of update method.
     */
    update(data, conditions, connection = null) {
        return this._checkRules(data)
            .then(() => {
                if (typeof this.beforeUpdate === "function") {
                    return this.beforeUpdate(data);
                }

                return Promise.resolve(data);
            })
            .then(values => this.appDatasource.update(this.useTable, values, conditions, connection));
    }

    /**
     * Use this to update many records in database.
     * @param {Array} data - List of data to be updated.
     * @param {Object} conditions - Object conditions.
     * @param {Object} connection - Database connection.
     * @returns {Object} - Result of saveMany method.
     */
    updateMany(data, conditions, connection = null) {
        return this._checkRules(data).then(() => this.appDatasource.updateMany(this.useTable, data, conditions, connection));
    }

    findBy(params, connection = null) {
        return this.appDatasource.findBy(this.useTable, this.getParams(params), connection);
    }

    findAll(params, connection = null) {
        return this.appDatasource.findAll(this.useTable, this.getParams(params), connection);
    }

    find(type, params, connection = null) {
        return this.appDatasource.find(this.useTable, type, this.getParams(params), connection);
    }

    getParams(data) {
        let params = data;

        if (this._join) {
            if (params) {
                params.join = this._join;
            } else {
                params = {
                    join: this._join,
                };
            }
        }

        return params;
    }

    showSQL() {
        this.appDatasource.showSQL();
    }

    getDatasource() {
        if (this.appDatasource.dbConfig.dataSource === "PgSQL") {
            return this.appDatasource.pgsql;
        } else if (this.appDatasource.dbConfig.dataSource === "MongoDB") {
            return this.appDatasource.mongoDB;
        }
    }
}

export default AppModel;
