/*eslint no-underscore-dangle: ["error", { "allow": ["_getWhereSQL", "_getSelectSQL", "_getCount"] }]*/
"use strict";

import DataSource from "../../src/Component/DataSource/PgSQL.js";
import { expect } from "chai";
let pgsql = null;

before(function () {
    const config = {
        dataSource: "MySQL",
        host: "127.0.0.1",
        port: 3306,
        user: "root",
        password: "<password>",
        database: "db",
    };

    pgsql = new DataSource(config);
});

describe("Testing the datasource pgsql.", function () {
    it("Should validate conditions array", function (done) {
        const params = {
            conditions: [
                {
                    name: "Diego",
                    operation: "<>",
                },
                {
                    perfil: "ADMINISTRATOR",
                    operation: "=",
                },
            ],
        };

        const res = pgsql._getWhereSQL("user", params);
        expect("WHERE user.name <> $1 AND user.perfil = $2").to.equal(res.where);
        expect("Diego").to.equal(res.values[0]);
        done();
    });

    it("Should validate conditions array with indexParam", function (done) {
        const params = {
            conditions: [
                {
                    name: "Diego",
                    operation: "<>",
                },
                {
                    perfil: "ADMINISTRATOR",
                    operation: "=",
                },
            ],
        };

        const res = pgsql._getWhereSQL("user", params, 2);
        expect("WHERE user.name <> $3 AND user.perfil = $4").to.equal(res.where);
        expect("Diego").to.equal(res.values[0]);
        done();
    });

    it("Should validate conditions object", function (done) {
        const params = {
            conditions: {
                name: "Diego",
                age: "10",
            },
        };

        const res = pgsql._getWhereSQL("user", params);
        expect("WHERE user.name = $1 AND user.age = $2").to.equal(res.where);
        expect("Diego").to.equal(res.values[0]);
        expect("10").to.equal(res.values[1]);
        done();
    });

    it("Should validate conditions object with indexParam", function (done) {
        const params = {
            conditions: {
                name: "Diego",
                age: "10",
            },
        };

        const res = pgsql._getWhereSQL("user", params, 2);
        expect("WHERE user.name = $3 AND user.age = $4").to.equal(res.where);
        expect("Diego").to.equal(res.values[0]);
        expect("10").to.equal(res.values[1]);
        done();
    });

    it("Should validate the _getSelectSQL function", function (done) {
        const params = {
            conditions: [
                {
                    name: "Diego",
                    operation: "<>",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE user.name <> $1").to.equal(res.sql);
        done();
    });

    it("Should validate fields in the _getSelectSQL function", function (done) {
        const params = {
            fields: ["name", "age"],
            conditions: [
                {
                    name: "Diego",
                    operation: "<>",
                },
                {
                    perfil: "ADMINISTRATOR",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT user.name,user.age FROM user WHERE user.name <> $1 AND user.perfil = $2").to.equal(res.sql);
        done();
    });

    it("Should validate the between condition.", function (done) {
        const params = {
            fields: ["name", "age"],
            conditions: [
                {
                    created_at: ["2019-01-01", "2019-12-01"],
                    operation: "BETWEEN",
                },
                {
                    perfil: "ADMINISTRATOR",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT user.name,user.age FROM user WHERE user.created_at BETWEEN $1 AND $2 AND user.perfil = $3").to.equal(res.sql);
        expect("2019-01-01").to.equal(res.values[0]);
        expect("2019-12-01").to.equal(res.values[1]);
        done();
    });

    it("Should validate the IS NOT NULL condition.", function (done) {
        const params = {
            fields: ["name", "age"],
            conditions: [
                {
                    name: "",
                    operation: "IS NOT NULL",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT user.name,user.age FROM user WHERE user.name IS NOT NULL").to.equal(res.sql);
        done();
    });

    it("Should validate the IS NULL condition.", function (done) {
        const params = {
            fields: ["name", "age"],
            conditions: [
                {
                    name: "",
                    operation: "IS NULL",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT user.name,user.age FROM user WHERE user.name IS NULL").to.equal(res.sql);
        done();
    });

    it("Should validate the IN condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: [1, 5, 10],
                    operation: "IN",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE user.id IN ($1, $2, $3)").to.equal(res.sql);
        done();
    });

    it("Should validate the OR condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: [1, 5, 10],
                    operation: "OR",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE  ((user.id = $1) OR (user.id = $2) OR (user.id = $3))").to.equal(res.sql);
        done();
    });

    it("Should validate the IN and OR condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: [1, 5, 10],
                    operation: "OR",
                },
                {
                    tag: ["tag-1", "tag-2", "tag-3"],
                    operation: "IN",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE  ((user.id = $1) OR (user.id = $2) OR (user.id = $3)) AND user.tag IN ($4, $5, $6)").to.equal(res.sql);
        done();
    });

    it("Should validate the AND, IN, OR condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: "10",
                },
                {
                    tag: ["tag-1", "tag-2", "tag-3"],
                    operation: "IN",
                },
                {
                    code: ["code-1", "code-2", "code-3"],
                    operation: "OR",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE user.id = $1 AND user.tag IN ($2, $3, $4) AND  ((user.code = $5) OR (user.code = $6) OR (user.code = $7))").to.equal(res.sql);
        done();
    });

    it("Should validate the AND, AND, OR condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: "10",
                },
                {
                    name: "Test",
                },
                {
                    code: ["code-1", "code-2", "code-3"],
                    operation: "OR",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE user.id = $1 AND user.name = $2 AND  ((user.code = $3) OR (user.code = $4) OR (user.code = $5))").to.equal(res.sql);
        done();
    });

    it("Should validate the AND, OR, AND condition.", function (done) {
        const params = {
            conditions: [
                {
                    id: "10",
                },
                {
                    code: ["code-1", "code-2", "code-3"],
                    operation: "OR",
                },
                {
                    name: "Test",
                },
            ],
        };
        const res = pgsql._getSelectSQL("user", params);
        expect("SELECT * FROM user WHERE user.id = $1 AND  ((user.code = $2) OR (user.code = $3) OR (user.code = $4)) AND user.name = $5").to.equal(res.sql);
        done();
    });

    it("Should return the sql count.", function (done) {
        const params = {
            conditions: {
                name: "Test",
            },
        };

        const res = pgsql._getCount("user", params);
        expect(res.sql).to.equal("SELECT COUNT(*) FROM user WHERE user.name = $1");
        done();
    });
});
